#include "math.h"
#include "gsl/gsl_multimin.h"
#include "gsl/gsl_vector.h"
#include "utils.h"
#include "kernel.h"

typedef struct  {
    ok_kernel* k;
    double** pars;
    double* steps;
} ok_dfmin_params;

double K_dfmin_f(const gsl_vector* cur, void* params) {
    ok_dfmin_params* dp = (ok_dfmin_params*) params;
    
    ok_kernel* k = dp->k;
    double** pars = dp->pars;
    
    for (int i = 0; i < cur->size; i++)
        *(pars[i]) = cur->data[i];
    
    k->flags |= NEEDS_SETUP;
    K_calculate(k);
    double res = k->minfunc(k);
    //printf("f %e %e %e\n", MGET(k->system->elements, 2, MASS), MGET(k->system->elements, 3, MASS), res);
    return res;
}

void K_dfmin_df(const gsl_vector* cur, void* params, gsl_vector* g) {
    ok_dfmin_params* dp = (ok_dfmin_params*) params;
    
    ok_kernel* k = dp->k;
    double** pars = dp->pars;
    double* steps = dp->steps;
    ok_callback mf = k->minfunc;
//    printf("df %e %e %e\n", MGET(k->system->elements, 2, MASS), MGET(k->system->elements, 3, MASS), 0.);
    for (int j = 0; j < cur->size; j++) {
        for (int i = 0; i < cur->size; i++) {
            *(pars[i]) = cur->data[i] - (i == j ? 0.5*steps[j] : 0.);
        }
        k->flags |= NEEDS_SETUP;
        K_calculate(k);
        double f1 = mf(k);
        
        for (int i = 0; i < cur->size; i++) {
            *(pars[i]) = cur->data[i] + (i == j ? 0.5*steps[j] : 0.);
        }
        k->flags |= NEEDS_SETUP;
        K_calculate(k);
        
        double f2 = mf(k);
        VSET(g, j, (f2-f1)/steps[j]);
        //printf("%e %e %e %e\n", cur->data[j], *(pars[j]), steps[j], (f2-f1)/steps[j]);
    }
}

void K_dfmin_fdf(const gsl_vector* cur, void* params, double* f, gsl_vector* g) {
    ok_dfmin_params* dp = (ok_dfmin_params*) params;
    
    ok_kernel* k = dp->k;
    double** pars = dp->pars;
    double* steps = dp->steps;
    ok_callback mf = k->minfunc;
    
    for (int i = 0; i < cur->size; i++) {
            *(pars[i]) = cur->data[i];
        }
    k->flags |= NEEDS_SETUP;
    K_calculate(k);
    *f = k->minfunc(k);
    //printf("fdf %e %e %e\n", MGET(k->system->elements, 2, MASS), MGET(k->system->elements, 3, MASS), *f);

    for (int j = 0; j < cur->size; j++) {
        
        for (int i = 0; i < cur->size; i++) {
            *(pars[i]) = cur->data[i] - (i == j ? 0.5*steps[j] : 0.);
        }
        k->flags |= NEEDS_SETUP;
        K_calculate(k);
        double f1 = k->minfunc(k);
        
        for (int i = 0; i < cur->size; i++) {
            *(pars[i]) = cur->data[i] + (i == j ? 0.5*steps[j] : 0.);
        }
        k->flags |= NEEDS_SETUP;
        K_calculate(k);
        
        double f2 = mf(k);
        VSET(g, j, (f2-f1)/steps[j]);
    }
}

int K_minimize_dfmin(ok_kernel* k, int maxiter, double params[]) {
    double eps = 2e-3;
    
    int i = 0;
    if (params != NULL) {
        while (true) {
            if (params[i] == DONE)
                break;
            else if (round(params[i]) == OPT_EPS)
                eps = params[i+1];
            i+=2;
        }
    }
    int npars = 0;
    
    // Count all the parameters to minimize on
    for (int i = 1; i < k->system->nplanets + 1; i++)
        for (int j = 0; j < ELEMENTS_SIZE; j++)
            npars += (MIGET(k->plFlags, i, j) & MINIMIZE ? 1 : 0);
    for (int i = 0; i < k->parFlags->size; i++)
        npars += (VIGET(k->parFlags, i) & MINIMIZE ? 1 : 0);

    // Create a pointer table (flat array -> matrices)
    double** pars = (double**) malloc(sizeof(double*) * npars);
    
    int idx = 0;
    gsl_vector* x = gsl_vector_alloc(npars);
    gsl_vector* step = gsl_vector_alloc(npars);
    
    for (int i = 1; i < k->system->nplanets + 1; i++)
        for (int j = 0; j < ELEMENTS_SIZE; j++)
            if (MIGET(k->plFlags, i, j) & MINIMIZE) {
                
                pars[idx] = gsl_matrix_ptr(k->system->elements, i, j);
                
                x->data[idx] = MGET(k->system->elements, i, j);
                step->data[idx] = MGET(k->plSteps, i, j);
                idx++;
            }

    for (int i = 0; i < k->parFlags->size; i++)
        if (VIGET(k->parFlags, i) & MINIMIZE) {
            pars[idx] = gsl_vector_ptr(k->params, i);
            x->data[idx] = VGET(k->params, i);
            step->data[idx] = VGET(k->parSteps, i);
            idx++;
        }
        
    ok_dfmin_params sp;
    sp.k = k;
    sp.pars = pars;
    sp.steps = step->data;
    
    // Initialize minimizer
    gsl_multimin_fdfminimizer* s = gsl_multimin_fdfminimizer_alloc(gsl_multimin_fdfminimizer_vector_bfgs2,
                npars);
    gsl_multimin_function_fdf fdf;
    fdf.n = npars;
    fdf.f = K_dfmin_f;
    fdf.df = K_dfmin_df;
    fdf.fdf = K_dfmin_fdf;
    fdf.params = &sp;
    
    gsl_multimin_fdfminimizer_set(s, &fdf, x, 1e-2, 0.1);
    
    
    int iter = 0;
    int status;
    
    // Loop until the desired eps (the size of the simplex) is reached.
    
    ok_progress pr = k->progress;
    double minimum = DBL_MAX;
    
    do {
        iter++;
        status = gsl_multimin_fdfminimizer_iterate(s);
        double min = gsl_multimin_fdfminimizer_minimum(s);
        
        if (status)
            break;
        
        if (fabs(minimum - min) < eps * fabs(minimum) && iter > 10)
            break;
        minimum = MIN(min, minimum);

        if (pr != NULL)
            if (pr(iter, maxiter, k, __func__) != PROGRESS_CONTINUE) {
                status = PROGRESS_STOP;
                break;
            }
        
    } while (iter < maxiter);
    
    gsl_multimin_fdfminimizer_free(s);
    free(pars);
    gsl_vector_free(step);
    gsl_vector_free(x);
    return status;
}

