/*
 *  kernel.h
 *  Systemic2
 *
 *
 */

#include "systemic.h"
#include "utils.h"

#define ACTIVE (1 << 1)
#define MINIMIZE (1 << 2)

#define NEEDS_COMPILE (1 << 1)
#define NEEDS_SETUP (1 << 2)
#define BOOTSTRAP_DATA (1 << 3)
#define RETAIN (1 << 4)
#define SHARE_FLAGS (1 << 5)
#define SHARE_STEPS (1 << 6)
#define SHARE_DATA (1 << 7)
#define FREED (1 << 8)
#define DUPLICATE (1 << 9)

#define SIMPLEX 0
#define DFMIN 1

extern double ok_default_steps[];

// MEMORY MANAGEMENT
// Allocates a new kernel
ok_kernel* K_alloc();
// Frees the memory associated with the kernel
void K_free(ok_kernel* k);
// Returns a new copy of the current kernel (that you manage)
ok_kernel* K_clone(ok_kernel* k);
// Returns a new copy of the current kernel (that you manage), with data sharing options
ok_kernel* K_cloneFlags(ok_kernel* k, unsigned int shareFlags);

// RV MANAGEMENT
// add a new rv from an ASCII file; returns NULL on error
gsl_matrix* K_addDataFile(ok_kernel* k, const char* path, int type);
// add a new rv from a matrix
gsl_matrix* K_addDataTable(ok_kernel* k, gsl_matrix* rvtable, const char* name, int type);
void K_removeData(ok_kernel* k, int idx);
gsl_matrix* K_getData(ok_kernel* k, int idx);

// returns a compiled view of the data
double** K_compileData(ok_kernel* k);
// returns a matrix of the residuals
gsl_matrix* K_getResiduals(ok_kernel* k);
// returns a matrix of the data suitable for the periodogram
gsl_matrix* K_combineData(ok_kernel* k, const int data_type);

// ELEMENTS MANAGEMENT
void K_addPlanet(ok_kernel* k, const double elements[]);
void K_removePlanet(ok_kernel* k, int idx);
void K_setElement(ok_kernel* k, int row, int col, double value);
double K_getElement(ok_kernel* k, int row, int col);
void K_setElements(ok_kernel* k, gsl_matrix* elements);
gsl_matrix* K_getElements(ok_kernel* k);
void K_setElementFlag(ok_kernel* k, int row, int col, int value);
int K_getElementFlag(ok_kernel* k, int row, int col);
void K_setElementStep(ok_kernel* k, int row, int col, double value);
double K_getElementStep(ok_kernel* k, int row, int col);
void K_setElementRange(ok_kernel* k, int row, int col, double min, double max);
void K_getElementRange(ok_kernel* k, int row, int col, double* min, double* max);
gsl_matrix* K_getAllElements(ok_kernel* k);

// STELLAR PARAMETERS
K_GETSET_H(Mstar, Mstar, double)
        
// SYSTEM PARAMETERS
K_GETSET_H(epoch, Epoch, double)
K_GET_H(chi2, Chi2, double)
K_GET_H(chi2_nr, Chi2_nr, double)
K_GET_H(rms, Rms, double)
K_GET_H(jitter, Jitter, double)
K_GET_H(compiled, Compiled, double**)
K_GETSET_H(flags, Flags, unsigned int)
K_GETSET_H(minfunc, MinFunc, ok_callback)
K_GETSET_H(plSteps, ElementSteps, gsl_matrix*)
K_GETSET_H(parSteps, ParSteps, gsl_vector*)
K_GETSET_H(plFlags, ElementFlags, gsl_matrix_int*)
K_GETSET_H(parFlags, ParFlags, gsl_vector_int*)

K_GETSET_H(intMethod, IntMethod, int)
K_GETSET_H(intOptions, IntOptions, ok_integrator_options*)

unsigned int K_getNplanets(ok_kernel* k);
unsigned int K_getNRV(ok_kernel* k);
unsigned int K_getNdata(ok_kernel* k);
void K_getRange(ok_kernel* k, double* from, double* to);



// PARAMS MANAGEMENT
void K_setPars(ok_kernel* k, gsl_vector* pars);
gsl_vector* K_getPars(ok_kernel* k);
void K_setPar(ok_kernel* k, int idx, double val);
double K_getPar(ok_kernel* k, int idx);
void K_setParFlag(ok_kernel* k, int idx, int value);
int K_getParFlag(ok_kernel* k, int idx);
void K_setParStep(ok_kernel* k, int idx, double value);
double K_getParStep(ok_kernel* k, int idx);
void K_setParRange(ok_kernel* k, int idx, double min, double max);
void K_getParRange(ok_kernel* k, int idx, double* min, double* max);

// Load/Save
bool K_save(ok_kernel* k, FILE* fid);
ok_kernel* K_load(FILE* fid, int skip);

// OPERATIONS
// Calculate chi^2
void K_calculate(ok_kernel* k);
int K_minimize(ok_kernel* k, int algo, int maxiter, double params[]);
int K_1dminimize(ok_kernel* k, int algo, int maxiter, int row, int column, double params[]);

ok_system** K_integrate(ok_kernel* k, gsl_vector* times, ok_system** bag);
ok_system** K_integrateRange(ok_kernel* k, double from, double to, unsigned int samples, ok_system** bag);
// DEBUG
void K_print(ok_kernel* k, FILE* f);
void K_save_old(ok_kernel* k, const char* stem);
void K_setSeed(ok_kernel* k, unsigned long int seed);