
#include "kl.h"
#include "kernel.h"
#include "integration.h"
#include "gsl/gsl_vector.h"
#include "gsl/gsl_statistics.h"
#include "gsl/gsl_sort_vector.h"

ok_list* KL_alloc(const int size) {
    ok_list* kl = (ok_list*) malloc(sizeof(ok_list));
    kl->kernels = (ok_kernel**) calloc(size, sizeof(ok_kernel*));
    kl->size = size;
    return kl;
}

void KL_free(ok_list* kl) {
    if (kl == NULL)
        return;
    
    for (int i = 0; i < kl->size; i++) {
        kl->kernels[i]->system->flag = FREED;
    }
        
    free(kl->kernels);
    free(kl);
}

ok_list* KL_load(FILE* fid) {
    char line[8192];
    int np = -1;
    int idx = -1;
    int trials = -1;
    double Mstar = 0.;
    double epoch = 0.;
    
    ok_list* kl = NULL;
    
    while (true) {
        long int ft = ftell(fid);
        if ((fgets(line, sizeof(line), fid) == NULL) || feof(fid))
            break;
        
        if (line[0] == '#') {
            char tag[100] = {0};
            sscanf(line + 2, "%s = ", tag);
            
            if (strcmp(tag, "Planets") == 0) {
                sscanf(line + 2, "%*s = %d", &np);
            } else if (strcmp(tag, "Trials") == 0) {
                sscanf(line + 2, "%*s = %d", &trials);
                kl = KL_alloc(trials);
            } else if (strcmp(tag, "Epoch") == 0) {
                sscanf(line + 2, "%*s = %le", &epoch);
            } else if (strcmp(tag, "Mstar") == 0) {
                sscanf(line + 2, "%*s = %le", &Mstar);
            }
            
        } else {
            fseek(fid, ft, SEEK_SET);
            for (int tr = 0; tr < trials; tr++) {
                ok_kernel* k = K_alloc();
                ok_free_system(k->system);
                k->system = ok_alloc_system(np);
                K_setEpoch(k, epoch);
                K_setMstar(k, Mstar);
                
                for (int i = 0; i < ELEMENTS_SIZE; i++) {
                    for (int j = 1; j <= np; j++) {
                        double v;
                        fscanf(fid, "%le", &v);
                        MSET(k->system->elements, j, i, v);
                    }
                }

                for (int i = 0; i < PARAMS_SIZE; i++) {
                    double v;
                    fscanf(fid, "%le", &v);
                    VSET(k->params, i, v);
                }

                fscanf(fid, "%le", &k->chi2);
                kl->kernels[tr] = k;
            }
            break;
        }
    }
    
    return kl;
}

gsl_vector* KL_getElements(const ok_list* kl, const int pl, const int el) {
    gsl_vector* v = gsl_vector_alloc(kl->size);
    
    for (int i = 0; i < kl->size; i++)
        VSET(v, i, K_getElement(kl->kernels[i], pl, el));
    return v;
}

gsl_vector* KL_getPars(const ok_list* kl, const int vo) {
    gsl_vector* v = gsl_vector_alloc(kl->size);
    
    for (int i = 0; i < kl->size; i++)
        VSET(v, i, K_getPar(kl->kernels[i], vo));
    
    return v;
}

gsl_matrix* KL_getElementsStats(const ok_list* kl, const int what) {
    
    gsl_vector* v = gsl_vector_alloc(kl->size);
    int npl = MROWS(kl->kernels[0]->system->elements);
    gsl_matrix* m = gsl_matrix_alloc(npl, ELEMENTS_SIZE);
    gsl_matrix_set_all(m, 0.);
    
    for (int i = 0; i < npl; i++)
            for (int j = 0; j < ELEMENTS_SIZE; j++) {
                for (int n = 0; n < kl->size; n++) {
                    VSET(v, n, MGET(kl->kernels[n]->system->elements, i, j));
                }

                switch (what) {
                    case STAT_MEAN:
                        if (j == MA || j == LOP || j == INC || j == NODE)
                            MSET(m, i, j, ok_average_angle(v->data, v->size, false));
                        else
                            MSET(m, i, j, gsl_stats_mean(v->data, 1, v->size));
                        break;
                    case STAT_STDDEV:
                        if (j == MA || j == LOP || j == INC || j == NODE) {
                            MSET(m, i, j, ok_stddev_angle(v->data, v->size, false));
                        }
                        else
                            MSET(m, i, j, gsl_stats_sd(v->data, 1, v->size));
                        break;
                    case STAT_MEDIAN:
                        if (j == MA || j == LOP || j == INC || j == NODE)
                            MSET(m, i, j, ok_median_angle(v->data, v->size, false));
                        else {
                            gsl_sort_vector(v);
                            MSET(m, i, j, gsl_stats_median_from_sorted_data(v->data, 1, v->size));
                        }
                        break;
                    default:
                        // percentiles
                        gsl_sort_vector(v);
                        MSET(m, i, j, gsl_stats_quantile_from_sorted_data(v->data, 1, v->size, (double)(what)/100.));
                };
            }
    gsl_vector_free(v);
    return m;
}

unsigned int KL_getSize(const ok_list* kl) {
    return kl->size;
}

void KL_removeAtIndex(ok_list* kl, const unsigned int idx) {
    K_free(kl->kernels[idx]);
    for (int i = idx + 1; i < kl->size; i++)
        kl->kernels[i-1] = kl->kernels[i];
    kl->size--;
}

void KL_fprintf(const ok_list* kl, FILE* out, const char* fmt, const char* lfmt) {
    lfmt = (lfmt != NULL ? lfmt : "%10s%d ");
    
    int np = kl->kernels[0]->system->nplanets + 1;
    int vo = PARAMS_SIZE;
    
    fprintf(out, "# Planets = %d\n", np);
    fprintf(out, "# Trials = %d\n", kl->size);
    fprintf(out, "# Mstar = %e\n", K_getMstar(kl->kernels[0]));
    fprintf(out, "# Epoch = %e\n", K_getEpoch(kl->kernels[0]));
    
    fprintf(out, "## ");
    
    for (int i = 0; i < ELEMENTS_SIZE; i++)
        for (int j = 1; j < np; j++)
            fprintf(out, lfmt, ok_orb_labels[i], j);
    for (int i = 0; i < vo; i++)
        fprintf(out, lfmt, "PARAM", i);
    fprintf(out, lfmt, "CHI2");
    fprintf(out, "\n");
    
    for (int m = 0; m < kl->size; m++) {
        ok_kernel* k = kl->kernels[m];
        for (int i = 0; i < ELEMENTS_SIZE; i++)
            for (int j = 1; j < np; j++)
                fprintf(out, fmt, MGET(k->system->elements, j, i));
        for (int i = 0; i < vo; i++)
            fprintf(out, fmt, VGET(k->params, i));
        fprintf(out, fmt, k->chi2);
        fprintf(out, " \n");
    }
    
}