#include "mcmc.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_randist.h"
#include "math.h"
#include "assert.h"

double K_mcmc_merit(ok_kernel* k) {
    K_calculate(k);
    double b = k->params->data[NOISE];
    double chi2 = K_getChi2(k);
    
    double prior = 1./b;
    for (int i = 1; i < MROWS(k->system->elements); i++) { 
        prior *= 1./MGET(k->system->elements, i, PER) * 1./MGET(k->system->elements, i, MASS);
    }
    
    return pow(b, 0.5 * k->ndata) * exp(-0.5*b*chi2) * prior;
}

ok_list* K_mcmc(ok_kernel* k, int steps, int discard, double params[], gsl_matrix** diags) {
    ok_list* kl = KL_alloc((int)(steps/discard));
    
    gsl_matrix* plSteps = k->plSteps;
    gsl_vector* parSteps = k->parSteps;
    gsl_matrix_int* plFlags = k->plFlags;
    gsl_vector_int* parFlags = k->parFlags;
    int nbodies = plSteps->size1;
    int nvo = parSteps->size;
    
    double acc = 0.;
    
    double prevMerit = K_mcmc_merit(k);
    ok_kernel* prevK = k;
    
    
    for (unsigned int i = 0; i < steps; i++) {
            ok_kernel* k2 = K_cloneFlags(prevK, SHARE_DATA | SHARE_FLAGS | SHARE_STEPS);
            k2->flags &= ~RETAIN;
            
            gsl_matrix* els = k2->system->elements;
            gsl_vector* pars = k2->params;
            
            for (int j = ELEMENTS_SIZE; j < nbodies * ELEMENTS_SIZE; j++) {
                if (plFlags->data[j] & MINIMIZE) {
                    els->data[j] += gsl_ran_gaussian(k->rng, plSteps->data[j]);
                }
            }
            for (int j = 0; j < nvo; j++) 
                if (parFlags->data[j] & MINIMIZE) {
                    pars->data[j] += gsl_ran_gaussian(k->rng, parSteps->data[j]); 
                }
            
            K_calculate(k2);
            double merit = K_mcmc_merit(k2);    
            assert(!isnan(merit));
            
            double al = MIN(merit/prevMerit, 1.);
            double u = gsl_rng_uniform(k->rng);
            if (u < al) {
                prevMerit = merit;
                if (!(prevK->flags & RETAIN) && (prevK != k)) {
                    K_free(prevK);
                    prevK = NULL;
                }
                acc += 1.;
                prevK = k2;
            }
            
            if (i % discard == 0) {
                if (prevK->flags & RETAIN || prevK == k)
                    kl->kernels[(int)(i/discard)] = K_cloneFlags(prevK, SHARE_DATA | SHARE_FLAGS | SHARE_STEPS);
                else
                    kl->kernels[(int)(i/discard)] = prevK;
                
                prevK->flags |= RETAIN;
            }
            
            if (! (k2->flags & RETAIN) && (prevK != k2)) {
                K_free(k2);
            }
            
            if (i % discard == 0) {
                printf("%d %e %e %e %e %d %e %e\n", i, merit, K_getElement(prevK, 1, PER),  K_getElement(prevK, 1, MASS), K_getElement(prevK, 1, ECC), 
                        kl->kernels[(int)(i/discard)]->flags, K_getPar(prevK, NOISE),
                        acc/(double)i);
            }
    }
    
    return kl;
}
